/* Socks Server 5
 * Copyright (C) 2002 - 2006 by Matteo Ricchetti - <matteo.ricchetti@libero.it>

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "SS5Main.h"
#include "SS5Mod_balance.h"


S5RetCode InitModule( struct _module *m )
{
  m->AddVip  = AddVip;
  m->AddConn = S5AddConn2Real;
  m->RemoveConn = S5RemoveConn2Real;
  m->LoadBalancing = LoadBalancing;
  m->Balancing = Balancing;
  m->FreeConnectionTable = FreeConnectionTable;
  m->FreeAffinity = FreeAffinity;

  return OK;
}

S5RetCode AddVip (char *real, unsigned int vid, unsigned int idx)
{
  S5AddReal2ConnectionTable(real, vid, idx);
 
  return OK;
}

S5RetCode FreeConnectionTable (struct _S5ConnectionEntry *ce)
{
  free(ce);
 
  return OK;
}

S5RetCode LoadBalancing( struct _SS5ClientInfo *ci, struct _SS5RequestInfo *ri )
{
  struct in_addr s;
  struct in_addr d;

  unsigned int vid;
  unsigned int ttl_status;

  /*
   *    If affinity enabled, looks for affinity between src ip and dst ip
   *    before using balancing
   */
  if( SS5SocksOpt.Sticky ) {
    IFLINUX( inet_aton(ci->SrcAddr,&s); )
    IFSOLARIS( inet_pton(AF_INET, ci->SrcAddr, &s); )

    ttl_status = OK;
    vid = S5GetRealVid(ri->DstAddr);

    LOCKMUTEXCA()
    if( (d.s_addr = S5GetAffinity(s.s_addr,&ttl_status,vid)) == ERR ) {
      if( ttl_status == ERR ) {
        /*
         *    If age expired, remove affinity between src ip and dst ip
         */
        S5RemoveAffinity(s.s_addr,vid);
      }

      /*
       *    Balances with least connections
       */
      S5LeastConnectionReal(ri->DstAddr);

      /*
       *    Setup affinity between src ip and dst ip
       */
      IFSOLARIS( inet_pton(AF_INET, ri->DstAddr, &d); )
      IFLINUX( inet_aton((const char *)ri->DstAddr,&d); )

      S5SetAffinity(s.s_addr,d.s_addr,vid);
    }
    else {
      strcpy(ri->DstAddr,inet_ntoa(d));
    }
    UNLOCKMUTEXCA()
  }
  else {
    /*
     *    Balances with least connections
     */
    S5LeastConnectionReal(ri->DstAddr);
  }
  
  return OK;
}

S5RetCode S5LeastConnectionReal(char *s5application)
{
  register unsigned int i;
  register unsigned int j;

  unsigned int id1;
  unsigned int vid;
  unsigned int conn;

  LOCKMUTEXCT()
  /*
   *    Search for one real with the smaller number of connections
   */
  for(i = 0; i < NReal; i++) {
    if( STREQ(S5ConnectionTable.S5ConnectionEntry[i]->Real,s5application,strlen(s5application)) ) {
      id1  = i;
      vid  = S5ConnectionTable.S5ConnectionEntry[i]->Vid; 
      conn = S5ConnectionTable.S5ConnectionEntry[i]->Connection;

      for(j = 0; j<NReal; j++)
        if( S5ConnectionTable.S5ConnectionEntry[j]->Vid == vid )
	  if(S5ConnectionTable.S5ConnectionEntry[j]->Connection < conn ) {
            conn = S5ConnectionTable.S5ConnectionEntry[j]->Connection;
	    id1  = j;
	  }
      strncpy(s5application,S5ConnectionTable.S5ConnectionEntry[id1]->Real,sizeof(S5ConnectionTable.S5ConnectionEntry[id1]->Real) - 1);
      UNLOCKMUTEXCT()
      return OK;
    }
  }
  UNLOCKMUTEXCT()
  return ERR;
}

S5RetCode S5AddReal2ConnectionTable(char *real, unsigned int vid, unsigned int idx)
{
  S5ConnectionTable._tmp_S5ConnectionEntry[idx] = (struct _S5ConnectionEntry *)calloc(1,sizeof(struct _S5ConnectionEntry));

  strncpy(S5ConnectionTable._tmp_S5ConnectionEntry[idx]->Real,real,strlen(real));
  S5ConnectionTable._tmp_S5ConnectionEntry[idx]->Vid = vid;
  S5ConnectionTable._tmp_S5ConnectionEntry[idx]->Connection = 0;

  return OK;
}

S5RetCode S5GetRealVid(char *real)
{
  register unsigned int idx;

  for(idx = 0; idx < NReal; idx++)
    if( STREQ(S5ConnectionTable.S5ConnectionEntry[idx]->Real,real,strlen(real)) ) {
      return S5ConnectionTable.S5ConnectionEntry[idx]->Vid;
    }
  return ERR;
}

S5RetCode S5AddConn2Real(char *real)
{
  register unsigned int idx;

  for(idx = 0; idx < NReal; idx++)
    if( STREQ(S5ConnectionTable.S5ConnectionEntry[idx]->Real,real,strlen(real)) ) {
      LOCKMUTEXCT()
      S5ConnectionTable.S5ConnectionEntry[idx]->Connection++;
      UNLOCKMUTEXCT()
      return OK;
    }
  return ERR;
}

S5RetCode S5RemoveConn2Real(char *real)
{
  register unsigned int idx;

  for(idx = 0; idx < NReal; idx++)
    if( STREQ(S5ConnectionTable.S5ConnectionEntry[idx]->Real,real,strlen(real)) ) {
      if( S5ConnectionTable.S5ConnectionEntry[idx]->Connection ) {
        LOCKMUTEXCT()
        S5ConnectionTable.S5ConnectionEntry[idx]->Connection--;
        UNLOCKMUTEXCT()
      }
      return OK;
    }
  return ERR;
}

inline S5Limit S5StickyHash( unsigned long int srcip )
{
  return (srcip % MAXSTICKYLIST);
}

unsigned long int S5GetAffinity(unsigned long int srcip, unsigned int *ttl_status, unsigned int vid)
{
  S5Limit idx;

  struct _S5StickyNode *node;

  idx = S5StickyHash( srcip );

  if( S5StickyList[idx] == NULL )
    return ERR;
  else {
    node = S5StickyList[idx];

    do {
     if( (node->srcip == srcip)  && (node->vid == vid)) {
       if( node->ttl > time(NULL) ) {
         return node->dstip;
       }
       else {
	 *ttl_status = ERR;
         return ERR;
       }
     }  
     node = node->next;
    } while(node != NULL );
  }
  return ERR;
}

S5RetCode S5SetAffinity(unsigned long int srcip, unsigned long int dstip, unsigned int vid )
{
  int idx;
  struct _S5StickyNode *node;
  struct in_addr s,d;

  s.s_addr = srcip;
  d.s_addr = dstip;

  idx=S5StickyHash( srcip );

  if( S5StickyList[idx]== NULL ) {
    S5StickyList[idx] = (struct _S5StickyNode *)calloc(1,sizeof(struct _S5StickyNode));
    S5StickyList[idx]->srcip = srcip;
    S5StickyList[idx]->dstip = dstip;
    S5StickyList[idx]->ttl = (time(NULL) + SS5SocksOpt.StickyAge);
    S5StickyList[idx]->vid = vid;
  }
  else {
    node=S5StickyList[idx];
    while( node->next != NULL ){
      node=node->next;
    }
    node->next = (struct _S5StickyNode *)calloc(1,sizeof(struct _S5StickyNode));
    node->next->srcip = srcip;
    node->next->dstip = dstip;
    node->next->ttl = (time(NULL) + SS5SocksOpt.StickyAge);
    node->next->vid = vid;
    node->next->next = NULL;
  }
  return OK;
}

S5RetCode S5RemoveAffinity(unsigned long int srcip, unsigned int vid)
{
  int idx;
  struct _S5StickyNode *node;
  struct _S5StickyNode **backnode;

  idx = S5StickyHash( srcip );

  node = S5StickyList[idx];
  backnode = &S5StickyList[idx];

  do {
   if( (node->srcip == srcip) && (node->vid == vid) ) {
     if( node->next != NULL )
       *backnode = node->next;
     else
       *backnode = NULL;

     free(node);
     break;
   }  
   backnode = &node->next;
   node = node->next;

  } while(node != NULL );

  return OK;
}

S5RetCode Balancing( struct _SS5ClientInfo *ci, struct _SS5Socks5Data *sd )
{

  register unsigned int idx;

  unsigned long int bufsize; 

  struct in_addr s;
  struct in_addr d;

  struct _S5StickyNode *node;

  time_t currentAge;

  char *httpResponse;
  char  buf[512];

  if( STREQ(sd->MethodRequest,"GET /balancing HTTP/1.",sizeof("GET /balancing HTTP/1.") - 1) 
     || STREQ(sd->MethodRequest,"GET /balancing/ HTTP/1.",sizeof("GET /balancing/ HTTP/1.") - 1) ) {
    /*
     *     Create http response
     */
    httpResponse = (char *)calloc(4096+256*NReal,sizeof(char));

    currentAge = time(NULL);

    snprintf(httpResponse,(4096+256*NReal),
"HTTP/1.1 200 OK\r\n\
Server: Socks Server 5-V3.0\r\n\
Connection: close\r\n\
Content-Type: text/html; charset=ISO-8859-1\r\n\
Content-Language: en\r\n\r\n\
<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\"> \
<html> \
<head> \
  <meta content=\"text/html; charset=ISO-8859-1\" \
 http-equiv=\"content-type\"> \
  <meta http-equiv=\"refresh\" content=\"60\"> \
  <title>SS5V3</title> \
</head> \
<body> \
<table style=\"width: 100%%; text-align: left;\" cellpadding=\"0\" \
 cellspacing=\"0\"> \
  <tbody> \
    <tr> \
      <td style=\"width: 100px; vertical-align: top;\"><img alt=\"\" \
 src=\"SS5Logo.jpg\" \
 style=\"height: 79px; width: 106px;\"><br> \
      </td> \
      <td style=\"vertical-align: top; text-align: right;\"><a \
 style=\"font-weight: bold; font-family: tahoma;\" href=\"http://ss5.sourceforge.net/\"><font \
 size=\"-1\"></font></a><a href=\"http://ss5.sourceforge.net/\"><font size=\"-1\"><span \
 style=\"font-family: tahoma;\">SS5 Home Page</span></font></a> <br> \
      <br> \
      <hr style=\"width: 100%%; height: 2px;\"><br> \
      </td> \
    </tr> \
  </tbody> \
</table> \
<br> \
<table \
 style=\"background-color: rgb(85, 111, 147); width: 100%%; text-align: left;\" \
 cellpadding=\"2\"> \
  <tbody> \
    <tr> \
      <td style=\"vertical-align: top;\"><font \
 style=\"color: rgb(255, 255, 255);\" size=\"-1\"><span \
 style=\"font-family: tahoma; font-weight: bold;\">SS5 Socks Server 5 \
Console</span></font><br> \
      </td> \
    </tr> \
  </tbody> \
</table> \
<table \
 style=\"background-color: rgb(220, 220, 220); width: 100%%; text-align: left;\" \
 cellpadding=\"2\"> \
  <tbody> \
    <tr> \
      <td style=\"vertical-align: top;\"><font \
 style=\"font-weight: bold; font-family: tahoma;\" size=\"-2\"> \
%s</font><br> \
      </td> \
      <td style=\"vertical-align: top; text-align: right;\"><font \
 size=\"-1\"><span style=\"font-family: tahoma;\">Join to SS5 </span><a \
 style=\"font-family: tahoma;\" \
 href=\"http://lists.penguin.it/mailman/listinfo/ss5\">Mailinglist</a></font><br> \
      </td> \
    </tr> \
  </tbody> \
</table> \
<br> \
<br> \
<table style=\"text-align: left; width: 987px; height: 92px;\" \
 cellpadding=\"2\"> \
  <tbody> \
    <tr style=\"font-family: tahoma;\"> \
      <td \
 style=\"height: 25px; background-color: rgb(173, 207, 206); vertical-align: top;\"><font \
 size=\"-1\"><span style=\"font-weight: bold;\">Balancing Table</span><br> \
      </font> </td> \
    </tr> \
    <tr style=\"font-family: tahoma;\" align=\"left\"> \
      <td style=\"vertical-align: top;\"> \
      <table \
 style=\"background-color: rgb(220, 220, 220); width: 100%%; text-align: left;\" \
 border=\"1\" cellpadding=\"2\" cellspacing=\"2\"> \
        <tbody> \
          <tr> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Destination \
address<br> \
            </font></td> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Vid<br> \
            </font></td> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Current \
connections</font></td> \
          </tr>",
     ctime(&currentAge));
 
  for(idx = 0; idx < NReal; idx++) {
    snprintf(buf,sizeof(buf) - 1,
"<tr> \
   <td style=\"vertical-align: top;\"><font size=\"-1\">%s<br> \
   </font></td> \
   <td style=\"vertical-align: top;\"><font size=\"-1\">%u<br> \
   </font></td> \
   <td style=\"vertical-align: top;\"><font size=\"-1\">%u<br> \
   </font> </td> \
 </tr>",S5ConnectionTable.S5ConnectionEntry[idx]->Real,
        S5ConnectionTable.S5ConnectionEntry[idx]->Vid,
        S5ConnectionTable.S5ConnectionEntry[idx]->Connection);

    strncat(httpResponse,buf,strlen(buf));
  }

  strncat(httpResponse,
"</tbody> \
      </table> \
      </td> \
    </tr> \
  </tbody> \
</table> \
<br> \
<table style=\"text-align: left; width: 987px; height: 91px;\" \
 cellpadding=\"2\"> \
  <tbody> \
    <tr> \
      <td \
 style=\"height: 25px; background-color: rgb(173, 207, 206); vertical-align: top;\"><font \
 size=\"-1\"><span style=\"font-family: tahoma; font-weight: bold;\">Affinity \
Table<br> \
      </span></font></td> \
    </tr> \
    <tr style=\"font-family: tahoma;\" align=\"left\"> \
      <td style=\"vertical-align: top;\"> \
      <table \
 style=\"background-color: rgb(220, 220, 220); width: 100%%; text-align: left;\" \
 border=\"1\" cellpadding=\"2\" cellspacing=\"2\"> \
        <tbody> \
          <tr> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Source \
address</font><font size=\"-1\"> </font></td> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Vid<br> \
            </font></td> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Destination \
address<br> \
            </font></td> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Sticky age<br> \
            </font> </td> \
            <td style=\"vertical-align: top;\"><font size=\"-1\">Current age<br> \
            </font></td> \
          </tr>\0",2048);

   bufsize=4096 + 256*NReal + 2048;

   for(idx = 0; idx < 997; idx++) {
          node = S5StickyList[idx];

          while( node != NULL ) {
            char cur_time[64];

            bufsize += 512;
            httpResponse = (char *)realloc(httpResponse,bufsize);

            s.s_addr = node->srcip;
            d.s_addr = node->dstip;

            currentAge = time(NULL);
#ifdef SOLARIS
            ctime_r(&currentAge,cur_time,sizeof(cur_time));
#else
            ctime_r(&currentAge,cur_time);
#endif

            snprintf(buf,sizeof(buf) - 1,
"<tr> \
    <td style=\"vertical-align: top;\"><font size=\"-1\">%s<br> \
    </font></td> \
    <td style=\"vertical-align: top;\"><font size=\"-1\">%u<br> \
    </font></td> \
    <td style=\"vertical-align: top;\"><font size=\"-1\">%s<br> \
    </font></td> \
    <td style=\"vertical-align: top;\"><font size=\"-1\">%s<br> \
    </font> </td> \
    <td style=\"vertical-align: top;\"><font size=\"-1\">%s<br> \
    </font></td> \
 </tr>",    inet_ntoa(s),node->vid,
            inet_ntoa(d),ctime(&node->ttl),
            cur_time);

            strncat(httpResponse,buf,strlen(buf));
            node = node->next;
          }
      }

      strncat(httpResponse,
"</tbody> \
 </table> \
 </td> \
 </tr> \
 </tbody> \
</table> \
<br> \
<br> \
</body> \
</html>",256);

    /*
     *    Send http response
     */
    if( send(ci->Socket,httpResponse,strlen(httpResponse),SS5_SEND_OPT) == -1) {
      free(httpResponse);
      return ERR;
    }
    else {
      fcntl(ci->Socket,F_SETFL,O_NONBLOCK);
      recv(ci->Socket,httpResponse,strlen(httpResponse),0);
      free(httpResponse);
      return OK;
    }
  }
  else if( STREQ(sd->MethodRequest,"GET /SS5Logo.jpg HTTP/1.",sizeof("GET /SS5Logo.jpg HTTP/1.") - 1) ||
           STREQ(sd->MethodRequest,"GET /balancing/SS5Logo.jpg HTTP/1.",sizeof("GET /balancing/SS5Logo.jpg HTTP/1.") - 1) ) {
    /*
     *    Create http response
     */
    httpResponse = (char *)calloc(256,sizeof(char));

    snprintf(httpResponse,256,
"HTTP/1.1 200 OK\r\n\
Server: Socks Server 5-V3.2\r\n\
Connection: close\r\n\
Content-Type: image/jpg\r\n\r\n");

    /*
     *    Send http response
     */
    if( send(ci->Socket,httpResponse,strlen(httpResponse),SS5_SEND_OPT) == -1) {
      free(httpResponse);
      return ERR;
    }
    else {
      send(ci->Socket,S5LogoBuffer,65536,SS5_SEND_OPT);
      fcntl(ci->Socket,F_SETFL,O_NONBLOCK);
      recv(ci->Socket,httpResponse,strlen(httpResponse),0);
      free(httpResponse);

      return OK;
    }
  }

  return ERR;
}

S5RetCode FreeAffinity( struct _S5StickyNode **node )
{
  struct _S5StickyNode *lnode;
  struct _S5StickyNode *lnode_prev=NULL;

  lnode=*node;

  if( lnode != NULL ) {
    do {
      while( lnode->next != NULL ) {
        lnode_prev=lnode;
        lnode=lnode->next;
      }
      free(lnode);
      if( lnode_prev != NULL ) {
        lnode_prev->next=NULL;
        lnode=lnode_prev;
        lnode_prev=NULL;
      }
      else
        lnode=NULL;
    } while( (lnode) != NULL );
  }
  *node=NULL;
  
  return OK;

}

